<?

/**
 * Campustream's notification system, backed by Redis
 */
class notifications {
	private $user_id;
	private $unread, $read;
	private $r;
	
	private static $instance;
	
	private static $types = array(
		'news',
		'newscomment',
		'collaborate',
		'reply'
	);
	
	/**
	 * Return a singleton instance of this class
	 */
	public static function load() {
		if (self::$instance == null) {
			self::$instance = new notifications();
		}

		return self::$instance;
	}
	
	/**
	 * Create a new notifications class and store a connection
	 * to Redis for later use.
	 */
	public function __construct($user=false) {
		if ($user instanceof User_Model) {
			$this->user_id = $user->id;
		} elseif (is_numeric($user)) {
			$this->user_id = $user;
		} else {
			$this->user_id = sess::getUserID();
		}
		
		$this->r = RedisManager::connection();
	}
	
	/**
	 * Retrieve a certain category of notifications.
	 *
	 * @param $type The category to retrieve
	 * @param $min The minimum number of notifications to return
	 */
	public function get($type, $min = 5) {
		$msgs = array();
		$len = $this->r->llen($this->getKey($type, 'unread'));
		if ($len > 0) {
			$temp = $this->r->lrange($this->getKey($type, 'unread'), 0, $len);
			foreach ($temp as $t) {
				$msgs[] = unserialize($t);
			}
		}
		
		if (count($msgs) < $min) {
			$temp = $this->r->lrange($this->getKey($type, 'read'), 0, ($min - count($msgs) - 1));
			foreach ($temp as $t) {
				$msgs[] = unserialize($t);
			}
		}
		
		return $msgs;
	}
	
	/**
	 * Send a notification to the given user.
	 *
	 * @param $to_user The user to send the notification to
	 * @param $type The category of notification
	 * @param $msg The notification message
	 * @param $link_id The ID that will be used to link the user to the post
	 */
	public function send($to_user, $type, $msg, $link_id) {
		// type = news | collaborate | reply
		$from = $this->user_id;
		
		$n = new notification($from, $type, $msg, $link_id);
		
		if ($to_user instanceof User_Model) {
			$to_user = $to_user->id;
		}
		
		$this->r->lpush("notifications:{$to_user}:{$type}:unread", serialize($n));
	}
	
	/**
	 * Mark all notifications as read
	 */
	public function markAllAsRead() {
		foreach (self::$types as $type) {
			while ($n = $this->r->rpop($this->getKey($type, 'unread'))) {
				$n = unserialize($n);
				$n->is_read = true;
				
				$this->r->lpush($this->getKey($type, 'read'), serialize($n));
			}
		}
	}
	
	/**
	 * Mark all notifications of a certain type as read
	 *
	 * @param $type The category of notifications
	 */
	public function markAsRead($type) {
		if (!in_array($type, self::$types)) {
			return false;
		}
		
		while ($n = $this->r->rpop($this->getKey($type, 'unread'))) {
			$n = unserialize($n);
			$n->is_read = true;
			
			$this->r->lpush($this->getKey($type, 'read'), serialize($n));
		}
		
		if ($type == 'news') {
			return $this->markAsRead('newscomment');
		}
		
		return true;
	}
	
	/**
	 * Retrieve the number of unread notifications, either in a certain
	 * category, or across all categories.
	 */
	public function numUnread($type=false) {
		$num = 0;
		
		if (!$type) {
			foreach (self::$types as $type) {
				$num += $this->r->llen($this->getKey($type, 'unread'));
			}
		} else {
			$num = $this->r->llen($this->getKey($type, 'unread'));
			
			if ($type == 'news') {
				$num += $this->r->llen($this->getKey('newscomment', 'unread'));
			}
		}
		
		
		return $num;
	}
	
	private function getKey($type, $status) {
		return "notifications:{$this->user_id}:$type:$status";
	}
}

/**
 * Class representing a single notification. This class is serialized
 * when it is saved in Redis, and unserialized when retrieved.
 */
class notification {
	public $from_user_id;
	public $message;
	public $time_sent;
	public $type;
	public $is_read;
	public $link_id;
	
	private $from_user;
	
	/**
	 * Creates a new notification
	 *
	 * @param $from Who is this notification from?
	 * @param $type What type of notification is this?
	 * @param $message The text content of the notification
	 * @param $link_id The ID to use in the link for this content
	 */
	public function __construct($from, $type, $message, $link_id) {
		if ($from instanceof User_Model) {
			$this->from_user_id = $from->id;
		} else {
			$this->from_user_id = $from;
		}
		
		$this->type = $type;
		$this->message = $message;
		$this->is_read = false;
		$this->time_sent = time();
		$this->link_id = $link_id;
	}
	
	private function get_user() {
		if (!$this->from_user) {
			$this->from_user = ActiveCache::find('User_Model', $this->from_user_id, 43200)->sql(
				"SELECT * FROM users WHERE id = {$this->from_user_id} LIMIT 1"
			);
		}
		
		return $this->from_user;
	}
	
	public function __get($key) {
		switch ($key) {
			case 'user': return $this->get_user();
		}
	}
}