<?

/**
 * Represents a single comment on a news/event item
 */
class Newscomment_Model extends ActiveRecord {
	public $table_name = 'newscomments';
	public $columns = array('id', 'posted_by', 'news_id', 'parent_id', 'content', 'time_posted', 'deleted');
	public $cache_keys = array( 'newscomment/#' => 'id', 'newscomment/user:#' => 'posted_by', 'newscomment/parent:#' => 'parent_id', 'newscomment/news:#' => 'news_id');
	public $has_many = array('children');
	public $has_one = array('parent_comment', 'user', 'news');
	public $public_columns = array( 'id', 'posted_by', 'news_id', 'parent_id', 'content', 'time_posted', 'children' );
	
	/**
	 * Recursively load the comment tree given a news/event ID
	 */
	public static function load_comment_tree($news_id) {
		$r = RedisManager::connection();
		$key = "news:{$news_id}:comment:0";
		$comments = array();
		
		$len = $r->llen($key);
		if ($len) {
			$ids = $r->lrange($key, 0, $len);
			
			foreach ($ids as $id) {
				$comment = ActiveCache::find('Newscomment_Model', $id, 43200)->sql(
					"SELECT * FROM newscomments WHERE id = $id LIMIT 1"
				);
				
				$comment->load_user();
				$comment->load_children(true);
				
				$comments[] = $comment;
			}
		}
		
		return $comments;
	}
	
	/**
	 * Load the user that posted this comment
	 */
	public function load_user() {
		if (!$this->user) {
			$user = ActiveCache::find('User_Model', $this->posted_by, 43200)->sql(
				"SELECT * FROM users WHERE id = {$this->posted_by} LIMIT 1"
			);
			
			if ($user->is_loaded()) {
				$this->add_relationship('user', $user);
			}
		}
	}
	
	/**
	 * Load the news item that this comment was posted under
	 */
	public function load_news() {
		if (!$this->news) {
			$news = ActiveCache::find('Newsevent_Model', $this->news_id, 43200)->sql(
				"SELECT * FROM newsevents WHERE id = {$this->news_id} LIMIT 1"
			);
			
			$this->add_relationship('news', $news);
		}
	}
	
	/**
	 * Load the parent comment of this comment (if any)
	 */
	public function load_parent() {
		if (!$this->parent_comment && $this->parent_id > 0) {
			$parent = ActiveCache::find('Newscomment_Model', $this->parent_id, 43200)->sql(
				"SELECT * FROM newscomments WHERE id = {$this->parent_id} LIMIT 1"
			);
			
			if ($parent->is_loaded() && $parent->deleted == 0) {
				$this->add_relationship('parent_comment', $parent);
			}
		}
	}
	
	/**
	 * Load all children comments of this comment
	 *
	 * @param $recursive Recursively load all comments below this one
	 */
	public function load_children($recursive = false) {
		$r = RedisManager::connection();
		$key = "news:{$this->news_id}:comment:{$this->id}";
		
		$len = $r->llen($key);
		if ($len) {
			$child_ids = $r->lrange($key, 0, $len);
			
			$children = array();
			foreach($child_ids as $child_id) {
				$child = ActiveCache::find('Newscomment_Model', $child_id, 43200)->sql(
					"SELECT * FROM newscomments WHERE id = $child_id LIMIT 1"
				);
				
				if ($child->is_loaded()) {
					$child->load_user();
					if ($recursive) {
						$child->load_children(true);
					}
				}
							
				$children[] = $child;
			}
			
			$this->add_relationship('children', $children);
		} else {
			$this->add_relationship('children', array());
		}
	}
	
	/**
	 * Save this comment and insert it into Redis as well as MySQL
	 */
	public function save($new = true) {
		parent::save();
		
		if (!$new) { return; }
		
		$r = RedisManager::connection();
		$key = "news:{$this->news_id}:comment:{$this->parent_id}";
		
		$r->rpush($key, $this->id);
		
		$this->load_news();
		$this->news->num_comments++;
		$this->news->save();
		
		$this->load_parent();
		if ($this->posted_by != $this->news->posted_by) {
			$n = notifications::load();
			$n->send($this->news->posted_by, 'news', "has posted a comment on your news/event: {$this->news->title}", $this->news->short_id);
			$this->news->load_user();
			
			email::send_notification($this->news->user, "has posted a comment on your news/event: {$this->news->title}", "/ne/{$this->news->short_id}");
		}
		
		if ($this->parent_comment && $this->parent_comment->posted_by != $this->posted_by) {
			$n->send($this->parent_comment->posted_by, 'newscomment', "has replied to your comment from: {$this->news->title}", $this->news->short_id);
			
			$this->parent_comment->load_user();
			email::send_notification($this->parent_comment->user, "has replied to your comment from: {$this->news->title}", "/ne/{$this->news->short_id}");
		}
	}
	
	/**
	 * Soft-delete this comment by flagging it as deleted
	 */
	public function delete() {
		$this->deleted = 1;
		$this->save(false);
	}
	
	public function __get($key) {
		switch ($key) {
			case 'childs': return $this->children;
		}
		
		return parent::__get($key);
	}
}