<?

/**
 * Represents a news/event item
 */
class Newsevent_Model extends ActiveRecord {
	public $table_name = 'newsevents';
	public $columns = array('id', 'type', 'category_id', 'newsevent', 'title', 'post', 'posted_by', 'time_posted', 'votes', 'num_comments');
	public $cache_keys = array( 'newsevent/#' => 'id', 'newsevents/type:#' => 'type', 'newsevents/category:#' => 'category', 'newsevents/user:#' => 'posted_by' );
	public $has_many = array('comments');
	public $has_one = array('category', 'user', 'eventmeta');
	public $public_columns = array( 'id', 'short_id', 'type', 'category_id', 'newsevent', 'title', 'post', 'posted_by', 'time_posted', 'votes', 'num_comments', 'comments', 'user', 'eventmeta' );
	
	/**
	 * Load the category that pertains to this news/event item
	 */
	public function load_category() {
		if (!$this->category) {
			$category = new Newseventcategory_Model();
			$category->load($this->category_id);
			
			$this->add_relationship('category', $category);
		}
	}
	
	/**
	 * Load the user that posted this news/event item
	 */
	public function load_user() {
		if (!$this->user) {
			$user = ActiveCache::find('User_Model', $this->posted_by, 43200)->sql(
				"SELECT * FROM users WHERE id = {$this->posted_by} LIMIT 1"
			);
			
			if ($user->is_loaded()) {
				$this->add_relationship('user', $user);
			}
		}
	}
	
	/**
	 * Load all comments posted under this news/event item
	 */
	public function load_comments() {
		if (!$this->comments && $this->num_comments > 0) {
			$comments = Newscomment_Model::load_comment_tree($this->id);
			$this->add_relationship('comments', $comments);
		}
	}
	
	/**
	 * Load metadata for this event
	 */
	public function load_eventmeta() {
		$this->load_category();
		
		if (!$this->eventmeta && $this->category->type == 'event') {
			$eventmeta = ActiveCache::find('Eventmeta_Model', "event:{$this->id}", 43200)->sql(
				"SELECT * FROM eventmeta WHERE event_id = {$this->id} LIMIT 1"
			);
			
			if ($eventmeta->is_loaded()) {
				$this->add_relationship('eventmeta', $eventmeta);
			}
		}
	}
	
	/**
	 * Load a list of popular news and/or events. Uses the Reddit algorithm
	 * for measuring popularity.
	 */
	public static function load_popular($opts = false) {
		$category = (isset($opts['cat_id']) ? $opts['cat_id'] : -1);
		$type = (isset($opts['type']) ? $opts['type'] : 'all');
		$page = (isset($opts['page']) ? $opts['page'] : 1);

		$key = "news/popular:$category/type:$type/page:$page";
		$query = "
		  SELECT
		  	id,
		  	(log(greatest(abs(votes), 1)) +
		  	(
		  		SIGN(votes)
		  		* (UNIX_TIMESTAMP(time_posted) - 1134028003)
		  	) / 45000.0) as rating
		  FROM newsevents";
		  
		  if ($category != -1 || $type != 'all') {
		  	$where = array();
		  	if ($category != -1) {
		  		$where[] = "category_id = $category";
		  	}
		  	if ($type != 'all') {
		  		$where[] = "newsevent = '$type'";
		  	}
		  	
		  	$query .= " WHERE " . implode(" AND ", $where);
		  }
		  
		$query .= "
			ORDER BY rating DESC
		  LIMIT 0, 20
		";
		
		$m = MemcacheManager::connection();
		$ids = $m->get($key);
		if ($m->getResultCode() == Memcached::RES_NOTFOUND) {
			$ids = array();
			
			$result = DatabaseManager::readConnection()->query($query);
			if ($result) {
				while ($data = $result->fetch_assoc()) {
					$ids[] = $data['id'];
				}
				
				$m->set($key, $ids, 300);
			}
		}
		
		$posts = array();
		foreach ($ids as $id) {
			$news = ActiveCache::find('Newsevent_Model', $id, 43200)->sql(
				"SELECT * FROM newsevents WHERE id = $id LIMIT 1"
			);
			
			if ($news->is_loaded()) {
				$posts[] = $news;
			}
		}
		
		return $posts;
	}
	
	/**
	 * Save this news/event item and post a status message about it
	 */
	public function save($voteup=false) {
		parent::save();
		
		if (!$voteup) {
			return;
		}
		
		$this->vote('up');
		
		// then create the status message for it
		$status = new Status_Model();
		$status->user_id = $this->posted_by;
		$status->source = 'site';
		$status->message = $this->title;
		$status->post_date = ActiveRecord::NOW();
		$status->type = 'newsevent';
		$status->content_id = $this->id;
		$status->is_public = 1;
		$status->save();
	}
	
	/**
	 * Add or remove a vote for this news/event item
	 */
	public function vote($dir) {
		// First we need to check to make sure this user hasn't voted on this response before
		$r = RedisManager::connection();
		$userID = sess::getUserID();
		$dir = strtolower($dir);
		$has_voted = $this->has_voted();
				
		if ($has_voted === 'up') {
			$this->votes--;
			$r->srem("news:{$this->id}:voting:up", $userID);
		} elseif ($has_voted === 'down') {
			$this->votes++;
			$r->srem("news:{$this->id}:voting:down", $userID);
		}
		
		$this->load_user();
		
		if ($has_voted === 'up' && $dir === 'up') {
			$this->save(false);
			$this->user->karma--;
			$this->user->save();
			return;
		}
		
		if ($has_voted === 'down' && $dir === 'down') {
			$this->save(false);
			$this->user->karma++;
			$this->user->save();
			return;
		}
		
		// New vote
		if ($dir === 'up') {
			$this->votes++;
			$this->user->karma++;
			$r->sadd("news:{$this->id}:voting:up", $userID);
		} else {
			$this->votes--;
			$this->user->karma--;
			$r->sadd("news:{$this->id}:voting:down", $userID);
		}
		
		$this->user->save();
		$this->save(false);
	}
	
	/**
	 * Has the given user (or the currently active user) voted on this
	 * news/event story yet?
	 */
	public function has_voted($user_id=false) {
		if (!$user_id) {
			$user_id = sess::getUserID();
		} elseif ($user_id instanceof User_Model) {
			$user_id = $user_id->id;
		}
		
		$r = RedisManager::connection();
		$up_vote = $r->sismember("news:{$this->id}:voting:up", $user_id);
		if (!$up_vote) {
			$down_vote = $r->sismember("news:{$this->id}:voting:down", $user_id);
			
			if ($down_vote) {
				return 'down';
			}
		} else {
			return 'up';
		}
		
		return false;
	}
	
	private function get_votes($which) {
		$r = RedisManager::connection();
		$num = $r->scard("news:{$this->id}:voting:$which");
		
		if ($num) {
		  return $num;
		} else {
		  return 0;
		}
	}
	
	public function __get($key) {
		switch ($key) {
			case 'short_id': return base_convert($this->id, 10, 36);
			case 'upvotes': return $this->get_votes('up');
			case 'downvotes': return $this->get_votes('down');
		}
		
		return parent::__get($key);
	}
}