<?

/**
 * Represents a single response to a question
 */
class Response_Model extends ActiveRecord {
	public $table_name = 'responses';
	public $columns = array('id', 'question_id', 'user_id', 'message', 'post_date', 'vote_count', 'is_public');
	public $cache_keys = array( 'response/#' => 'id', 'responses/question:#' => 'question_id', 'responses/user:#' => 'user_id' );
	public $has_one = array('question', 'user');
	public $public_columns = array( 'id', 'question_id', 'user_id', 'message', 'post_date', 'vote_count', 'up_votes', 'down_votes' );
	
	/**
	 * Load the user that posted this response
	 */
	public function load_user() {
		if (!$this->user) {
			$user = ActiveCache::find('User_Model', $this->user_id, 43200)->sql(
				"SELECT * FROM users WHERE id = {$this->user_id} LIMIT 1"
			);
			
			if ($user->is_loaded()) {
				$this->add_relationship('user', $user);
			}
		}
	}
	
	/**
	 * Load the corresponding question that this response was posted under
	 */
	public function load_question() {
		if (!$this->question) {
			$question = ActiveCache::find('Question_Model', $this->question_id, 43200)->sql(
				"SELECT * FROM questions WHERE id = {$this->question_id} LIMIT 1"
			);
			
			if ($question->is_loaded()) {
				$this->add_relationship('question', $question);
			}
		}
	}
	
	/**
	 * Add or remove a vote for this response
	 *
	 * @param $dir up or down
	 */
	public function vote($dir) {
		// First we need to check to make sure this user hasn't voted on this response before
		$r = RedisManager::connection();
		$userID = sess::getUserID();
		$dir = strtolower($dir);
		$has_voted = $this->has_voted();
				
		if ($has_voted === 'up') {
			$this->vote_count--;
			$r->srem("voting:response:up:{$this->id}", $userID);
		} elseif ($has_voted === 'down') {
			$this->vote_count++;
			$r->srem("voting:response:down:{$this->id}", $userID);
		}
		
		$this->load_user();
		
		if ($has_voted === 'up' && $dir === 'up') {
			$this->save(false);
			$this->user->karma--;
			$this->user->save();
			return;
		}
		
		if ($has_voted === 'down' && $dir === 'down') {
			$this->save(false);
			$this->user->karma++;
			$this->user->save();
			return;
		}
		
		if ($dir === 'up') {
			$this->vote_count++;
			$this->user->karma++;
		} else {
			$this->vote_count--;
			$this->user->karma--;
		}
		
		$this->user->save();
		
		// Add user ID to the voted set
		$r->sadd("voting:response:{$dir}:{$this->id}", $userID);
		$this->save(false);
	}
	
	/**
	 * Has the active user voted on this response yet?
	 */
	public function has_voted() {
		$r = RedisManager::connection();
		$userID = sess::getUserID();
		$has_voted = $r->sismember("voting:response:up:{$this->id}", $userID);
		if ($has_voted) {
			return 'up';
		}
		
		$has_voted = $r->sismember("voting:response:down:{$this->id}", $userID);
		if ($has_voted) {
			return 'down';
		}
		
		return false;
	}
	
	private function individual_vote_count($dir) {
		$r = RedisManager::connection();
		return $r->scard("voting:response:{$dir}:{$this->id}");
	}
	
	/**
	 * Save this response and post a status message
	 */
	public function save($post_status = true) {
		// We need to save this first, because it will change
		// 2 => 1 when we save to MySQL
		$is_public = $this->is_public;
		if ($is_public === 2) {
			$this->is_public = 0;
		}
		
		parent::save();

		if (!$post_status) {
			return true;
		}
		
		$this->load_question();
		if ($this->user_id != $this->question->user_id) {
			$n = notifications::load();
			$n->send($this->question->user_id, 'collaborate', "has posted a reply to your question: {$this->question->title}", $this->question->short_id);
		}
		
		$m = MemcacheManager::connection();
		$m->delete("responses/question:{$this->question_id}");
		
		if ($is_public == 2) {
			return true;
		}
		
		// then create the status message for it
		$status = new Status_Model();
		$status->user_id = $this->user_id;
		$status->source = 'site';
		$status->message = $this->message;
		$status->post_date = ActiveRecord::NOW();
		$status->type = 'response';
		$status->content_id = $this->question->id;
		$status->is_public = $this->is_public;
		$status->save();
		
	}
	
	public function __get($key) {
		switch ($key) {
			case 'up_votes': return $this->individual_vote_count('up');
			case 'down_votes': return $this->individual_vote_count('down');
		}
		
		return parent::__get($key);
	}
}