<?

/**
 * Represents a single status update
 */
class Status_Model extends ActiveRecord {

	public $table_name = 'statuses';
	public $columns = array('id', 'user_id', 'source', 'message', 'post_date', 'type', 'content_id', 'deleted', 'is_public');
	public $cache_keys = array( 'status/#' => 'id', 'statuses/user:#' => 'user_id');
	public $has_one = array('meta', 'content');
	public $has_many = array();
	public $public_columns = array( 'id', 'user_id', 'source', 'message', 'post_date', 'type', 'content_id', 'user', 'comments', 'num_comments', 'meta', 'content' );
	
	/**
	 * The max number of statuses to store per stream
	 */
	public static $list_max_length = 200;
	
	/**
	 * Save this status update in MySQL and Redis
	 */
	public function save($new_status = true) {		
		// First we save to MySQL
		parent::save();
		
		if (!$new_status) { return true; }
		
		$this->load_user();
		$this->user->current_status = $this->id;
		$this->user->save();
		
		if ($this->is_public) {
			stream::publish(stream::STREAM_PUBLIC, $this->type, $this->id);
		}
			
		stream::publish(stream::STREAM_FRIENDS, $this->type, $this->id);
		stream::publish(stream::STREAM_USER, $this->type, $this->id);
		
		if ($this->type == 'text') {
			$r = RedisManager::connection();
			
			if ($this->user->has_twitter() && $this->user->twitter()->export_enabled) {
				$data = array(
					'user' => $this->user_id,
					'payload' => $this->message
				);
				
				$r->rpush('queue:twitter:statuses', serialize($data));
			}
			
			if ($this->user->has_facebook() && $this->user->facebook()->export_enabled) {
				$data = array(
					'user' => $this->user_id,
					'payload' => $this->message
				);
				
				$r->rpush('queue:facebook:statuses', serialize($data));
			}
		}
	}
	
	/**
	 * Soft-delete this status update for data collection purposes
	 */
	public function delete() {
		// soft-delete
		$this->deleted = 1;
		$this->save(false);
		
		$r = RedisManager::connection();
		
		$last_status = 0;
		if ($r->llen("statuses:user_timeline:{$this->user_id}") > 0) {
			$last_status = reset($r->lrange("statuses:user_timeline:{$this->user_id}", 0, 1));
		}
		
		$this->load_user();
		$this->user->current_status = $last_status;
		$this->user->save();
		
		ActiveCache::invalidate_keys_for($this);
		
		return true;
	}
	
	/**
	 * Load the user who posted this status update
	 */
	public function load_user() {
		if (!$this->user) {
			$user = ActiveCache::find('User_Model', $this->user_id, 43200)->sql(
				"SELECT * FROM users WHERE id = {$this->user_id} LIMIT 1"
			);
			
			if ($user->is_loaded()) {
				$this->add_relationship('user', $user);
			}
		}
	}
	
	/**
	 * Load all comments posted under this status
	 */
	public function load_comments() {
		if (!$this->comments) {
			$comments = ActiveCache::find_all('Statuscomment_Model', "status:{$this->id}", 43200)->sql(
				"SELECT * FROM statuscomments WHERE status_id = {$this->id} ORDER BY id ASC"
			);
			
			if (count($comments) > 0) {
				$this->add_relationship('comments', $comments);
			}
		}
	}
	
	/**
	 * Load metadata relating to this status update
	 */
	public function load_meta() {
		if (!$this->meta) {
			$model = ucfirst($this->type) . '_Model';
			$table = $this->type . 's';
			$meta = ActiveCache::find($model, $this->content_id, 43200)->sql(
				"SELECT * FROM {$table} WHERE id = {$this->content_id} LIMIT 1"
			);
			
			if ($meta->is_loaded()) {
				$this->add_relationship('meta', $meta);
			}
		}
	}
	
	/**
	 * Load the content relating to this status update
	 */
	public function load_content() {
		if ($this->type == 'text' || !$this->content_id) { return; }
		
		if (!$this->content) {
			$model = null;
			$table = null;
			
			switch ($this->type) {
				case 'question':
					$model = 'Question_Model';
					$table = 'questions';
					break;
				case 'response':
					$model = 'Response_Model';
					$table = 'responses';
					break;
			}
			
			$content = ActiveCache::find($model, $this->content_id, 43200)->sql(
				"SELECT * FROM $table WHERE id = {$this->content_id} LIMIT 1"
			);
			
			if ($content->is_loaded()) {
				if ($this->type == 'response') {
					$content->load_question();
					$content->question->load_tags();
					$content->question->load_responses();
					$content->question->load_user();
				} elseif ($this->type == 'question') {
					$content->load_responses();
					$content->load_tags();
				}
				
				$this->add_relationship('content', $content);
			}
		}
	}
	
	private function num_comments() {
		$this->load_comments();
		return count($this->comments);
	}
	
	public function __get($key) {
		switch ($key) {
			case 'num_comments': return $this->num_comments();
			case 'short_id': return base_convert($this->id, 10, 36);
		}
		
		return parent::__get($key);
	}

}