<?php

class ActiveRecord {
	
	public static $NOW = null;
	public static $NULL = null;
	
	// member variables are prefixed in order to prevent polluting the
	// class space and conflicting w/ column names
	protected $_data          = null;
	protected $_relationships = array();
	
	protected $_dirty         = false;
	protected $_new           = true;
	protected $_orig_pk       = null;
	
	protected $_subscribers   = array();
  
	// these should all get set by the child class.
	// also, these would ideally be static variables but php does
	// not support late-static binding til 5.3, so it would not work as expected
	public $table_name     = null;
	public $columns        = null;
	public $primary_key    = 'id';
	public $primary_key_value  = null;
	public $auto_increment = true;
	public $has_many       = array();
	public $has_one        = array();
	public $belongs_to     = array();
	
	const _UPDATE = 'UPDATE `%s` SET %s WHERE %s;';
	const _INSERT = 'INSERT INTO `%s` (%s) VALUES (%s);';
	const _SELECT = 'SELECT * FROM `%s` WHERE %s LIMIT 1';
	const _DELETE = 'DELETE FROM `%s` WHERE %s LIMIT 1';
	
	// model_to_table_name( 'User_Model' ) -> users
	public static function model_to_table_name( $class ) {
		return strtolower( str_replace( "_Model", "", $class ) ) . 's';
	}
	
	// table_to_model_name( 'users' ) -> User_Model
	public static function table_to_model_name( $class ) {
		$class = self::depluralize($class);
		return ucfirst( $class ) . "_Model";
	}
	
	public static function depluralize($word){
	    // Here is the list of rules. To add a scenario,
	    // Add the plural ending as the key and the singular
	    // ending as the value for that key. This could be
	    // turned into a preg_replace and probably will be
	    // eventually, but for now, this is what it is.
	    //
	    // Note: The first rule has a value of false since
	    // we don't want to mess with words that end with
	    // double 's'. We normally wouldn't have to create
	    // rules for words we don't want to mess with, but
	    // the last rule (s) would catch double (ss) words
	    // if we didn't stop before it got to that rule. 
	    $rules = array( 
	        'ss' => false, 
	        'os' => 'o', 
	        'ies' => 'y', 
	        'xes' => 'x', 
	        'oes' => 'o', 
	        'ies' => 'y', 
	        'ves' => 'f',
	        'nses' => 'nse',
	        'ses' => 's', 
	        's' => '');
	    // Loop through all the rules and do the replacement. 
	    foreach(array_keys($rules) as $key){
	        // If the end of the word doesn't match the key,
	        // it's not a candidate for replacement. Move on
	        // to the next plural ending. 
	        if(substr($word, (strlen($key) * -1)) != $key) 
	            continue;
	        // If the value of the key is false, stop looping
	        // and return the original version of the word. 
	        if($key === false) 
	            return $word;
	        // We've made it this far, so we can do the
	        // replacement. 
	        return substr($word, 0, strlen($word) - strlen($key)) . $rules[$key]; 
	    }
	    return $word;
	}
	
	// return a unique token that will be replaced at query time with the NOW() function.
	public static function NOW() {
		if ( self::$NOW === null ) {
			self::$NOW = uniqid( "RESERVED_", true );
		}
		return self::$NOW;
	}

	public static function NULL( $test=null ) {
		
		if ( $test !== null ) {
			return $test;
		}
		
		if ( self::$NULL === null ) {
			self::$NULL = uniqid( "RESERVED_", true );
		}
		return self::$NULL;
	}
	
	// passthrough to mysqli->prepare, allows for using prepared statements
	// with the find method. safer and better than manually escaping data.
	// EXAMPLE:
	// $stmt = ActiveRecord::prepare( "SELECT * FROM users where username=?" );
	// $stmt->bindParam( "s", $value );
	// ActiveRecord::find( 'User', $stmt );
	public static function prepare( $clause ) {  
	  return DatabaseManager::readConnection()->prepare( $clause );
	}
	
	public static function find_all( $class, $clause, $pk_hint='id' ) {
		
		$result = self::find( $class, $clause, $pk_hint );

		if ( ! is_array( $result ) && $result->is_loaded() === true ) {
			$result = array( $result );
		} else if ( ! is_array( $result ) && $result->is_loaded() === false ) {
			// this is needed to clear out the array if find() returns a conveninent empty
			// object.
			$result = array();
		}
		
		return $result;
		
	}
	
	// $class   = the ActiveRecord childclass that you want returned
	// $clause  = the SQL for the statement or primary key clause
	// $pk_hint = the primary key OR array or primary keys
	// returns the object or an array of objects  
	public static function find( $class, $clause, $pk_hint='id' ) {

		if ( empty( $clause ) ) {
			return new $class();
		}

		$connection = DatabaseManager::readConnection();
		Logger::debug( "activerecord find $clause" );
		// figure out if we are doing a lookup by primary key, or
		// raw sql. assumes a primary key will always be numeric
		if ( is_numeric( $clause ) ) {
			$primary_key_clause = "`$pk_hint` = '$clause'";
			$sql = sprintf( self::_SELECT, ActiveRecord::model_to_table_name( $class ), $primary_key_clause );
			$stmt = $connection->query( $sql );
		} else if ( $clause instanceof MySQLi_STMT ) {
			$clause->execute();
			$stmt = $clause;
		} else {
			$stmt = $connection->query( $clause );
		}
		
		if ( $connection->errno !== 0 ) {
			throw new Exception( 'an error occured: ' . $connection->error );
		} 
		
		// return an empty object if the result set is empty
		if ( $stmt->num_rows === 0 ) return new $class();
		
		$fields = $stmt->fetch_fields();
		
		$table_map = array();
		foreach( $fields as $field ) {
			isset( $table_map[$field->table] ) ? null : $table_map[$field->table] = array();
			$table_map[$field->table][] = $field->name;
		}
		
		$result_set = array();
		$seen_pks   = array_fill_keys( array_keys( $table_map ), array() ); // prime the array to have table names set
				
		// loop over each row. we have to use fetch_row because
		// fetch_assoc has overlap problems when a join has the
		// same column name.
		while ( $row = $stmt->fetch_row() ) {
			$table_offset = 0; // this is how deep we can go into the $row array before we hit the next table
			$parent_pk  = null;

			foreach( $table_map as $table_name => $column_list ) {
				$column_count = count( $column_list );
				// join the column name with the data, basically we are taking fetch_row and fetch_fields and making our own fetch_assoc
				$column_data  = array_combine( $column_list, array_slice( $row, $table_offset, $column_count ) );
				
				if ( is_array( $pk_hint ) ) {
					$pk_name = $pk_hint[ $model_name ];
				} else {
					$pk_name = $pk_hint;
				}
				
				$pk_value = $column_data[ $pk_name ];
				
				// protects against left joins- we want to ignore any joins that have a null primary key
				// and avoid creating an empty object.
				if ( $pk_value === null ) continue;
				
				// get the model name of the table we are working on
				$model_name = ActiveRecord::table_to_model_name( $fields[ $table_offset ]->table );
				
				$table_offset += $column_count;
								
				if ( $model_name === $class ) {
					
					// we have seen this item before, skip it
					if ( isset( $seen_pks[ $table_name ][ $pk_value ] ) ) {
						$parent_pk = $column_data[ $pk_name ];
						continue;
					}
			
					$object = new $model_name( $column_data );
					$seen_pks[ $table_name ][ $pk_value ] = true;
					$result_set[ $pk_value ] = $object;
					
					// parent_pk is used to track the primary key of the 'main' object
					// for the join parts of this sql row
					$parent_pk = $pk_value;
					
				} else {
					if ( isset( $seen_pks[ $table_name ][ $parent_pk ][ $pk_value ] ) ) {
						continue;
					}

					$object = new $model_name( $column_data );
					
					$seen_pks[$table_name][$parent_pk][ $pk_value ] = true;
					
					$result_set[$parent_pk]->add_relationship( $model_name, $object );
					
				}	
			}
		}
		
		// if there is only one item in the result set, just return the front item. array_values is
		// used to re-order the array, since $result_set uses the value of the primary key as the array key.
		return count($result_set) == 1 ? reset($result_set) : array_values($result_set);
	}
	
	// $data = an array of initial data to load the object with.
	// if $data is null, it assumes you want to create a new/blank object
	public function __construct( $data=null ) {
		
		// we can figure out the table name, based on the child class.
		// we assume tables are lower case plural versions of the class name i.e, User -> users
		if ( $this->table_name === null ) {
			$this->table_name = ActiveRecord::model_to_table_name( get_class( $this ) );
		}

		$s_relation = array_merge( $this->belongs_to, $this->has_one );
		foreach( $s_relation as &$value ) {
			$value = strtolower( $value );
		}
		
		
		$s_relation = array_fill_keys( $s_relation, null );
		$p_relation = array_fill_keys( $this->has_many, array() );
				
		$this->_relationships = array_merge( $s_relation, $p_relation );
		
		if ( $data !== null ) { 
			$this->_new  = false;
			$this->_data = $data; // we trust that this will be correct

			// make a copy of the primary key, incase it ever gets changed
			if ( isset( $this->_data[ $this->primary_key ] ) ) {
				$this->_orig_pk = $this->_data[ $this->primary_key ];
				$this->primary_key_value = $this->_data[ $this->primary_key ];
			}
		} else { 
			$this->_data = array_fill_keys( $this->columns, null );
		}
		
	}
	
	// delete this object from the database
	public function delete( $skip_delete=false) {
		
		if ( $skip_delete === false ) {
			$where_clause = "`{$this->primary_key}`='{$this->_orig_pk}'";
			$delete_sql = sprintf( self::_DELETE, $this->table_name, $where_clause );
			// uncommented line below to re-enable deletions
			DatabaseManager::writeConnection()->query( $delete_sql );
		}
		ActiveCache::invalidate_keys_for( $this );
		$this->notify_subscribers();
		
	}
	
	// return true if this object has been saved
	public function is_saved() { return (! (bool)$this->_dirty); }
	// return true if this object loaded from the database
	public function is_loaded() { return (! $this->_new); }
	
  // used in validation. we define it here because if no validation method is
  // present, validation should always succeed 
	public function validate() { return true; }
	
  // used to embed objects which have a relationship to this class- i.e, joins
  // $class  = the class name of the object
  // $object = the object to embed
	public function add_relationship( $class, $object ) {

		$plural = ActiveRecord::model_to_table_name( $class );
				
		if ( in_array( $plural, $this->has_many ) ) {
			$this->_relationships[$plural][] = $object;
		} else {
			$singular = substr( ActiveRecord::model_to_table_name( $class ), 0, -1 );
			$this->_relationships[$singular] = $object;
		}
 	}
	
  // used to update or insert the data, depending on if this is a new object or
  // not
	public function save() {
		// only execute sql if the state has changed
		if ( $this->_new === true ) {
			$this->_insert();
		} else if ( $this->_dirty !== false ) {
			$this->_update();
		}
		
		$this->notify_subscribers();
	}
	
	public function subscribe( $subscriber ) {
		$this->_subscribers[] = $subscriber;
	}
	
	public function notify_subscribers() {
		foreach( $this->_subscribers as $subscriber ) {
			$subscriber->notify( $this );
		}
	}
	
	// internal method, it runs an update on the dirty columns
	private function _update() {
		
		$connection = DatabaseManager::writeConnection();
		
		// generate the where clause, WHERE primary_key = primary_key_value
		$where_clause = "`{$this->primary_key}`='{$this->_orig_pk}'";
		
		// generate the set list
		$update_cols = array();

		foreach( $this->_dirty as $column => $value ) {
			if ( $value === ActiveRecord::NOW() ) {
				$update_cols[] = "`$column`=NOW()";
				$this->_data[$column] = date ("Y-m-d H:i:s"); // approximate the actual value, so it's in the obj w/out re-reading from mysql
			} elseif ( $value === ActiveRecord::NULL() ) {
				$update_cols[] = "`$column`=NULL";
				$this->_data[$column] = null;
				continue;
			} else {
				$value = $connection->real_escape_string( $value );
				$update_cols[] = "`$column`='{$value}'";
			}
		}
		
		// convert the array to a comma delimited list
		$update_cols = join( ', ', $update_cols );
		
		// glue the sql together
		$update_sql = sprintf( self::_UPDATE, $this->table_name, $update_cols, $where_clause );
		Logger::debug( "updating $update_sql");

		$connection->query( $update_sql );
		
		if ( $connection->errno !== 0 ) {
			throw new Exception( 'an error occured: ' . $connection->error . ' ' . $update_sql );
		}
		
		$this->_dirty = false;
		
	}
	
	// internal method, it inserts any added data into the database as a new row
	private function _insert() {

		$connection = DatabaseManager::writeConnection();
		
		// lets make a list of all the columns, excluding primary_key
		// make a copy of the data, we need to wrap it in backticks
		    $data = $this->_data;
				$data = array();
				foreach ( $this->_data as $key => $value) {
					if ( $key == $this->primary_key && $this->auto_increment === true ) continue;
					
					if ( $value === ActiveRecord::NULL() ) {
						$data[$key] = "NULL";
						$this->_data[$key] = null;
						continue;
					}
					
					if ( $value === ActiveRecord::NOW() ){ 
						$data[$key] = "NOW()"; 
						$this->_data[$key] = date ("Y-m-d H:i:s"); // approximate NOW() for this object. will use real NOW() in the database.
						continue;
					}
					$value = $connection->real_escape_string( $value );
					$data[$key] = "'$value'";
				}

		    // join( ',', array_map( function($i){ return "'$i'";}, $data ) );
		    $insert_data = join( ',', $data );

		    // now lets make a list of all the columns, excluding primary_key
		    $column_list = array();

		    foreach( $this->columns as $column ) {
		      // skip the primary key row if it's auto increment
		      if ( $column == $this->primary_key && $this->auto_increment === true ) continue;
		      $column_list[] = "`$column`";
		    }

		    $column_list = join( ',', $column_list );

		    $insert_sql = sprintf( self::_INSERT, $this->table_name, $column_list, $insert_data );
				Logger::debug( "inserting $insert_sql");
			
				$stmt = $connection->query($insert_sql);
			
		    // if there is an autoincrement pk, lets grab it and set the value
		    if ( $this->primary_key !== null && $this->auto_increment === true ) {
					$pk_value = $connection->insert_id;
					$this->__set( $this->primary_key, $pk_value);
					$this->_orig_pk = $pk_value;
					$this->primary_key_value = $pk_value;
		    }

		    $this->_dirty = false;
		    $this->_new   = false;

		  }	
// php magic method that is run when you call isset on any orm-property
	public function __isset( $key ) {
		return isset( $this->_data[$key] );
	}
	
	// php magic method that is run when you retrieve any orm-property or relationship
	public function __get( $key ) {
		if ( isset( $this->_data[$key] ) ) {
			return $this->_data[$key];
		} else if ( isset( $this->_relationships[$key] ) ) {
			return $this->_relationships[$key];
		}
		return null;
	}
	
	// php magic method that is used for setting an orm-property
	public function __set( $key, $value ) {
		if ( array_key_exists( $key, $this->_data ) && $this->_data[$key] !== $value ) {
			$this->_data[$key] = $value;

			// mark the key as dirty
			$this->_dirty = ( $this->_dirty === false ? array() : $this->_dirty );
			$this->_dirty[$key] = $value;
		}
	}
	
	// php magic method that is called when the class is serialized- i.e, before
	// being put in memcache
	public function __sleep() {
		if ( $this->is_loaded ) {
			$this->save();
		}
		return array( '_data', '_relationships', '_orig_pk', 'table_name', 'columns',
		'primary_key', 'primary_key_value', 'auto_increment', 'has_many', 'has_one', 'belongs_to', '_new' );
	}
	
	// php magic method that is called when the class is unserialized- i.e, after
	// being retrieved from memcache
	public function __wakeup() {
	}
	
	// this is a mess. TODO: FIX IT!
	// returns a 'limited' version of the current AR object. this means that columns we want hidden
	// are not visible. condensed flag sets whether we name array elements or not. (xml needs all keys to
	// be set while json does not)
	public function limited_object( $condensed=true) {
	
    	foreach( $this->public_columns as $column ) {
			if ( isset( $this->_relationships[$column] ) ) continue;
    	  	$limited_obj->{$column} = $this->$column;
    	}
    	
    	foreach( $this->_relationships as $class_name => $object ) {
				
			if ( ! in_array($class_name, $this->public_columns) ) continue;
					
			if ( empty($object) ) { $limited_obj->{$class_name} = null; continue; }
			
			if ( is_array($object) ) {
							
				if ( $condensed === true ) {
    				$limited_obj->{"{$class_name}"} = array();
				} else {
					$limited_obj->{"{$class_name}"}->{substr($class_name, 0, -1)} = array();
				}
							
				foreach( $object as $ob ) {
					if ( empty( $ob ) ) continue;
					if ( $condensed === true ) {
						$limited_obj->{"{$class_name}"}[] = $ob->limited_object();
					} else {
						// with xml, we can't just put things into a nameless array. need to name every key.
						$limited_obj->{$class_name}->{substr($class_name, 0, -1)}[] = $ob->limited_object(false);
					}
				}
			} else {
    	    
				if ( ! $object->is_loaded() ) { $limited_obj->{$class_name} = null; continue; }
				$limited_obj->{$class_name} = $object->limited_object();
			}
		}
    	
		return $limited_obj;
	}
  
	// convert this object to a public json object
  public function to_json() {
    
    $limited_object = $this->limited_object();
    
    return json_encode($limited_object);
  }
  
  public function to_jsonp() {
  	
  	$limited_object = $this->limited_object();
  	
  	if(isset($_REQUEST['callback'])) {
  		$callback = $_REQUEST['callback']; 
  	} else {
  		$callback = 'TwitpicLoad';
  	}
  	
  	return $callback.'('.json_encode($limited_object).');';
  	
  }
  
	// convert this object to a public xml object
  public function to_xml() {
    
    require_once 'XML/Serializer.php';
    
    $limited_object = $this->limited_object(false);
    
    if ( substr( $this->table_name, -2 ) === 'es' ) {
		$class = substr( $this->table_name, 0, -2 );
	} elseif ( substr( $this->table_name, -1 ) === "s" ) {
		$class = substr( $this->table_name, 0, -1 );
	} else {
		$class = $this->table_name;
	}
    
    $options = array( "indent"    => "    ",
                      "linebreak" => "\n",
											"mode" => "simplexml",
                      "typeHints" => false,
                      "addDecl"   => true,
                      "encoding"  => "UTF-8",
                      "rootName"   => $class,
                      "defaultTagName" => "images" );

    $serializer = new XML_Serializer( $options );
    $serializer->serialize( $limited_object );
    return $serializer->getSerializedData();
  }

	public static function convert_rs($format, $array, $root, $tag) {
		if ( $format === 'json' ) {
			return self::rs_to_json($array,$root);
		} else {
			return self::rs_to_xml($array,$root,$tag);
		}
	}

	public static function rs_to_json($array, $root) {
		
		$rs = $array['rs'];
		unset($array['rs']);
		
		
		foreach($rs as $object) {
			$array[$root][] = $object->limited_object();
		}
		
		echo json_encode($array);
	}
	
	public static function rs_to_xml($array, $root, $tag) {
		
		$rs = $array['rs'];
		unset($array['rs']);
		
		
		foreach($rs as $object) {
			$array[] = $object->limited_object();
		}
    
    	$options = array( "indent"    => "    ",
    	                  "linebreak" => "\n",
    	                  "typeHints" => false,
    	                  "addDecl"   => true,
    	                  "encoding"  => "UTF-8",
    	                  "rootName"   => $root,
    	                  "defaultTagName" => $tag );
		
    	$serializer = new XML_Serializer( $options );
    	$serializer->serialize( $array );
    	return $serializer->getSerializedData();
	}
	
		
}